#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include "stack.h"

//testuje zda se nahodou nenachazime mimo pole 0:ano 1:ne
#define TEST(x,y) ((x<sea.xs && x>=0 && y<sea.ys && y>=0)?1:0)
//to same, ale s opacnym poradim parametru
#define REVTEST(x,y) ((x<sea.ys && x>=0 && y<sea.xs && y>=0)?1:0)
//pouze zkratka
#define NAVRAT if (navrat==1) return 1; if (navrat==2) return 2
//taktez
#define HIDDEN sea.array[prvek.y][prvek.x].hidden
//opet
#define MINE sea.array[prvek.y][prvek.x].mine

typedef struct seafield {
    unsigned int count:4;
    unsigned int mine:1;
    unsigned int hidden:1;
    unsigned int flag:1;
} SEAFIELD;

typedef struct sea {
    SEAFIELD **array;
    int xs;
    int ys;
    int mines;
} SEA;

//funkcni prototypy
//popis jednotlivych funkci je v api.h
int sea_malloc(SEA *sea);
void sea_free(SEA *sea);
void sea_fill(SEA sea, SEAFIELD fill);
void sea_randomize(SEA sea);
void sea_count_mines(SEA sea);
int sea_load(SEA *sea, char *file);
void sea_hide(SEA sea);
int sea_flag(SEA sea, int x, int y);
int sea_uncover(SEA sea, int x, int y);
int sea_uncover_block(SEA sea, int x, int y);
void sea_uncover_mines(SEA sea);
void sea_flag_mines(SEA sea);

int sea_malloc(SEA *sea) {
  int i;
  sea->array=(SEAFIELD **)malloc(sizeof(SEAFIELD *)*sea->ys);
  if (sea->array==NULL) return -1;

  for(i=0;i<sea->ys;i++) {
    sea->array[i]=(SEAFIELD *)malloc(sizeof(SEAFIELD)*sea->xs);
    if (sea->array[i]==NULL) return -1;
  }

  return 0;
}

void sea_free(SEA *sea) {
  int i;
  for (i=0;i<sea->ys;i++) free(sea->array[i]);
  free(sea->array);
}

void sea_fill(SEA sea, SEAFIELD fill) {
  int i, j;
  for (i=0;i<sea.ys;i++) {
    for (j=0;j<sea.xs;j++) {
      sea.array[i][j].count = fill.count;
      sea.array[i][j].mine = fill.mine;
      sea.array[i][j].hidden = fill.hidden;
      sea.array[i][j].flag = fill.flag;
    }
  }
}

void sea_randomize(SEA sea) {
  int i,j;
  srand(time(NULL));
  int pocet = sea.mines;
  while(pocet>0) {
    i=rand()%sea.ys;
    j=rand()%sea.xs;
    if (sea.array[i][j].mine==1) continue;
    else {
      sea.array[i][j].mine=1;
      pocet--; 
    }
  }
}

void sea_count_mines(SEA sea) {
  int i,j;
  for(i=0;i<sea.ys;i++) {
    for(j=0;j<sea.xs;j++) {
      if (REVTEST(i-1,j-1)) {
        if (sea.array[i-1][j-1].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i-1,j)) {
        if (sea.array[i-1][j].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i-1,j+1)) {
        if (sea.array[i-1][j+1].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i,j-1)) {
        if (sea.array[i][j-1].mine==1) sea.array[i][j].count++; 
      }
      if (REVTEST(i,j+1)) {
        if (sea.array[i][j+1].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i+1,j-1)) {
        if (sea.array[i+1][j-1].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i+1,j)) {
        if (sea.array[i+1][j].mine==1) sea.array[i][j].count++;
      }
      if (REVTEST(i+1,j+1)) {
        if (sea.array[i+1][j+1].mine==1) sea.array[i][j].count++;
      }
    }
  }
}

int sea_load(SEA *sea, char *file) {
  int i,j,k;
  FILE *soubor;
  char radek[38];
  soubor = fopen(file, "r");
  if (soubor==NULL) return -1;
  i=j=k=0;
  while(fgets(radek, 37, soubor)!=NULL) {	//cte po radcich
    while(radek[j]!='\0') {
      if (radek[j]=='1') { sea->array[i][j].mine = 1; k++; }
      if (radek[j]=='0') sea->array[i][j].mine = 0;
      j++;					//inkrementace sloupce
    }
    j = 0;					//nulovani sloupce
    i++;					//inkrementace radku
  }
  sea->mines = k;
  return 0;
}

void sea_hide(SEA sea) {
  int i,j;
  for(i=0;i<sea.ys;i++) for(j=0;j<sea.xs;j++) sea.array[i][j].hidden=1;
}

int sea_flag(SEA sea, int x, int y) {
  if (TEST(x,y)==0) return -3;
  if (sea.array[y][x].hidden==0) return -1;
  else {
    if (sea.array[y][x].flag == 1) sea.array[y][x].flag = 0;
    else sea.array[y][x].flag = 1;
  }
  return 0;
}

int sea_uncover(SEA sea, int x, int y) {
  static int odkryte;			  //pocita pocet odkrytych policek
  if (!TEST(x,y)) return -3;
  if (sea.array[y][x].flag==1) return -2; 
  if (sea.array[y][x].hidden==0) {
    return -1;
  }
  else {
    if (sea.array[y][x].mine==1) {
      return 1;
    }
    else {                                //neni mina, nenulovy count
      if (sea.array[y][x].count!=0) {
        sea.array[y][x].hidden=0;
        odkryte++;
      }
      else {                              //vubec nic - odkryvame okolni
        STACK *zasobnik = (STACK *)malloc(sizeof(STACK));
        XY prvek;
        prvek.x = x; prvek.y = y;
        if (stack_init(zasobnik)==0) {
          if (stack_push(zasobnik, prvek)==1) {
            return 1;
          }
          else {
            while(stack_empty(*zasobnik)==0) {
              if (stack_pop(zasobnik, &prvek)==1) break;;
              if (HIDDEN==1) { HIDDEN = 0; odkryte++; }
              if (sea.array[prvek.y][prvek.x].count==0) {
                prvek.x--; prvek.y--;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
                prvek.x++;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0) 
                  stack_push(zasobnik, prvek);
                prvek.x++;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0) 
                  stack_push(zasobnik, prvek);
                prvek.y++;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
                prvek.y++; 
		if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
                prvek.x--;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
                prvek.x--;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
                prvek.y--;
                if (TEST(prvek.x,prvek.y) && HIDDEN==1 && MINE==0)
                  stack_push(zasobnik, prvek);
              }
            }
          }
        }
      } 
    }
  }

  if (odkryte==(sea.xs*sea.ys-sea.mines)) return 2;
  return 0; 
}

int sea_uncover_block(SEA sea, int x, int y) {
  int navrat;
  if (TEST(x,y)==0) return -3;
  navrat=sea_uncover(sea, x-1, y-1); NAVRAT;
  navrat=sea_uncover(sea, x,   y-1); NAVRAT;
  navrat=sea_uncover(sea, x+1, y-1); NAVRAT;
  navrat=sea_uncover(sea, x-1, y);   NAVRAT;
  navrat=sea_uncover(sea, x,   y);   NAVRAT;
  navrat=sea_uncover(sea, x+1, y);   NAVRAT;
  navrat=sea_uncover(sea, x-1, y+1); NAVRAT;
  navrat=sea_uncover(sea, x,   y+1); NAVRAT;
  navrat=sea_uncover(sea, x+1, y+1); NAVRAT;
  return 0;
}

void sea_uncover_mines(SEA sea) {
  int i,j;
  for(i=0;i<sea.ys;i++) {
    for(j=0;j<sea.xs;j++) {
      sea.array[i][j].flag = 0;
      if (sea.array[i][j].hidden==1)
        sea.array[i][j].hidden=(sea.array[i][j].mine==1?0:1);      
    }
  }
}

void sea_flag_mines(SEA sea) {
  int i,j;
  for(i=0;i<sea.ys;i++) {
    for(j=0;j<sea.xs;j++) {
      sea.array[i][j].flag=(sea.array[i][j].mine==1?1:0);
    }
  }
}
